/*
 *  Copyright (C) 2006-2015 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ffmpegBackend.h"

#ifdef HAVE_FFMPEG

#include "loadsym.h"

#define CREDITS		"ffmpeg"
#define LINK		"http://www.ffmpeg.org/"

const char ffmpegBackend::name[]="Ffmpeg";

#ifdef _WIN32
# define AVFORMATLIB	"avformat-" AV_STRINGIFY(LIBAVFORMAT_VERSION_MAJOR) ".dll"
# define AVCODECLIB	"avcodec-" AV_STRINGIFY(LIBAVCODEC_VERSION_MAJOR) ".dll"
# define AVUTILLIB	"avutil-" AV_STRINGIFY(LIBAVUTIL_VERSION_MAJOR) ".dll"
#else
# define AVFORMATLIB	"libavformat.so." AV_STRINGIFY(LIBAVFORMAT_VERSION_MAJOR)
# define AVCODECLIB	"libavcodec.so." AV_STRINGIFY(LIBAVCODEC_VERSION_MAJOR)
# define AVUTILLIB	"libavutil.so." AV_STRINGIFY(LIBAVUTIL_VERSION_MAJOR)
#endif

const AutoDLL ffmpegBackend::_avformat(AVFORMATLIB);
const AutoDLL ffmpegBackend::_avcodec(AVCODECLIB);
const AutoDLL ffmpegBackend::_avutil(AVUTILLIB);

#if LIBAVFORMAT_VERSION_MAJOR < 54
int (*ffmpegBackend::dl_av_open_input_file)(AVFormatContext**, const char*, AVInputFormat*, int, AVFormatParameters*)=0;
void (*ffmpegBackend::dl_av_close_input_file)(AVFormatContext*)=0;
int (*ffmpegBackend::dl_av_find_stream_info)(AVFormatContext*)=0;
#else
int (*ffmpegBackend::dl_avformat_open_input) (AVFormatContext**, const char*, AVInputFormat*, AVDictionary**)=0;
void (*ffmpegBackend::dl_avformat_close_input)(AVFormatContext**)=0;
int (*ffmpegBackend::dl_avformat_find_stream_info)(AVFormatContext*, AVDictionary**)=0;
#endif
#if LIBAVCODEC_VERSION_MAJOR < 54
int (*ffmpegBackend::dl_avcodec_open)(AVCodecContext*, AVCodec*)=0;
int (*ffmpegBackend::dl_avcodec_decode_audio3)(AVCodecContext*, int16_t*, int*, AVPacket*)=0;
#else
int (*ffmpegBackend::dl_avcodec_open2)(AVCodecContext*, const AVCodec*, AVDictionary**)=0;
int (*ffmpegBackend::dl_avcodec_decode_audio4)(AVCodecContext*, AVFrame*, int*, const AVPacket*)=0;
AVFrame* (*ffmpegBackend::dl_av_frame_alloc)();
void (*ffmpegBackend::dl_av_frame_free)(AVFrame**)=0;
int (*ffmpegBackend::dl_av_sample_fmt_is_planar)(enum AVSampleFormat)=0;
int (*ffmpegBackend::dl_av_samples_get_buffer_size)(int*, int, int, enum AVSampleFormat, int)=0;
#endif
int (*ffmpegBackend::dl_av_read_frame)(AVFormatContext*, AVPacket*)=0;
int (*ffmpegBackend::dl_av_seek_frame)(AVFormatContext*, int, int64_t, int)=0;
AVDictionaryEntry* (*ffmpegBackend::dl_av_dict_get)(AVDictionary*, const char*, const AVDictionaryEntry*, int)=0;
#if LIBAVCODEC_VERSION_MAJOR < 55
AVCodec* (*ffmpegBackend::dl_avcodec_find_decoder)(enum CodecID)=0;
#else
AVCodec* (*ffmpegBackend::dl_avcodec_find_decoder)(enum AVCodecID)=0;
#endif
void (*ffmpegBackend::dl_av_init_packet)(AVPacket*)=0;
void (*ffmpegBackend::dl_avcodec_flush_buffers)(AVCodecContext*)=0;
int (*ffmpegBackend::dl_avcodec_close)(AVCodecContext*)=0;
void (*ffmpegBackend::dl_av_free_packet)(AVPacket*)=0;
// int64_t (*ffmpegBackend::dl_av_rescale_q)(int64_t, AVRational, AVRational)=0;

FXString ffmpegBackend::_ext="*.(";

// Based on code from GNUsound 0.7.4 - Copyright (C) 2002-2004, Pascal Haakmat
// http://www.gnu.org/software/gnusound/
// and on info from Martin Böhme's article "Using libavformat and libavcodec"
// http://www.inb.uni-luebeck.de/~boehme/using_libavcodec.html

/*****************************************************************/

size_t ffmpegBackend::fillBuffer(void* buffer, const size_t bufferSize, const unsigned int seconds) {

	size_t n=_decodeBufOffset;
	FXuchar* buf=_decodeBuf;

	while (n<=bufferSize) {
		while (!_packet.data) {
			if (dl_av_read_frame(_formatContext, &_packet)<0) {
				FXTRACE((10, "Last frame: %d\n", static_cast<int>(n)));
				memcpy(buffer, buf, n);
				_decodeBufOffset=0;
				return n;
			}
			if (_packet.stream_index!=_audioStreamIndex) {
				dl_av_free_packet(&_packet);
				_packet.data=0;
			} else
				_packetOffset=0;
		}
		int frame_size=MAX_AUDIO_FRAME_SIZE*2-n;

		AVPacket avpkt;
		dl_av_init_packet(&avpkt);
		avpkt.data=_packet.data+_packetOffset;
		avpkt.size=_packet.size-_packetOffset;
#if LIBAVCODEC_VERSION_MAJOR < 54
		const int used=dl_avcodec_decode_audio3(_audioStream->codec, (FXshort*)(buf+n), &frame_size, &avpkt);
#else
		AVFrame *frame = dl_av_frame_alloc();
		int got_frame=0;
		const int used=dl_avcodec_decode_audio4(_audioStream->codec, frame, &got_frame, &avpkt);
		if (used >= 0 && got_frame) {
			int plane_size;
			int data_size = dl_av_samples_get_buffer_size(&plane_size, _audioStream->codec->channels,
				frame->nb_samples,
				_audioStream->codec->sample_fmt, 1);
			if (frame_size < data_size) {
				FXTRACE((50, "output buffer size is too small for the current frame: (%d < %d)\n", frame_size, data_size));
				return 0;
			}

			if (_planar && _audioStream->codec->channels > 1) {
				// Interleave channels
				uint8_t *out=(uint8_t*)(buf+n);
				int idx=0;
				const int samples=plane_size/_sampleSize;

				for (int j=0; j<samples; j++) {
					for (int ch=0; ch<_audioStream->codec->channels; ch++) {
						memcpy(out, frame->extended_data[ch]+idx, _sampleSize);
						out+=_sampleSize;
					}
					idx+=_sampleSize;
				}
			} else {
				memcpy(buf+n, frame->extended_data[0], plane_size);
			}
			frame_size = data_size;
		} else {
			frame_size = 0;
		}
                dl_av_frame_free(&frame);
#endif
		FXTRACE((50, "frame_size: %d\n", frame_size));
		if (used>=0) {
			n+=frame_size;
			_packetOffset+=used;
		}
		if ((used<0) || (_packet.size<=_packetOffset)) {
			dl_av_free_packet(&_packet);
			_packet.data=0;
		}
	}

	memcpy(buffer, buf, bufferSize);
	_decodeBufOffset=n-bufferSize;
	memcpy(buf, buf+bufferSize, _decodeBufOffset);

	return bufferSize;
}

/*****************************************************************/

bool ffmpegBackend::init() {

	if (!_avformat.loaded() || !_avcodec.loaded() || !_avutil.loaded())
		return false;

#if LIBAVCODEC_VERSION_MAJOR < 54
	LOADSYM(_avcodec, avcodec_decode_audio3, int(*)(AVCodecContext*, int16_t*, int*, AVPacket*))
	LOADSYM(_avcodec, avcodec_open, int(*)(AVCodecContext*, AVCodec*))
#else
        LOADSYM(_avcodec, avcodec_open2, int(*)(AVCodecContext*, const AVCodec*, AVDictionary**))
	LOADSYM(_avcodec, avcodec_decode_audio4, int(*)(AVCodecContext*, AVFrame*, int*, const AVPacket*))
        LOADSYM(_avutil, av_frame_alloc, AVFrame*(*)())
        LOADSYM(_avutil, av_frame_free, void(*)(AVFrame**))
	LOADSYM(_avutil, av_sample_fmt_is_planar, int(*)(enum AVSampleFormat))
	LOADSYM(_avutil, av_samples_get_buffer_size, int(*)(int*, int, int, enum AVSampleFormat, int))
#endif

#if LIBAVFORMAT_VERSION_MAJOR < 54
	LOADSYM(_avformat, av_open_input_file, int(*)(AVFormatContext**, const char*, AVInputFormat*, int, AVFormatParameters*))
	LOADSYM(_avformat, av_close_input_file, void(*)(AVFormatContext*))
	LOADSYM(_avformat, av_find_stream_info, int(*)(AVFormatContext*))
#else
	LOADSYM(_avformat, avformat_open_input, int(*)(AVFormatContext **ps, const char *filename, AVInputFormat *fmt, AVDictionary **options))
	LOADSYM(_avformat, avformat_close_input, void(*)(AVFormatContext**))
	LOADSYM(_avformat, avformat_find_stream_info, int(*)(AVFormatContext*, AVDictionary**))
#endif
	LOADSYM(_avformat, av_read_frame, int(*)(AVFormatContext*, AVPacket*))
	LOADSYM(_avformat, av_seek_frame, int(*)(AVFormatContext*, int, int64_t, int))
	LOADSYM(_avutil, av_dict_get, AVDictionaryEntry*(*)(AVDictionary*, const char*, const AVDictionaryEntry*, int))
	LOADSYM(_avcodec, av_init_packet, void(*)(AVPacket*))
#if LIBAVCODEC_VERSION_MAJOR < 55
	LOADSYM(_avcodec, avcodec_find_decoder, AVCodec*(*)(enum CodecID))
#else
	LOADSYM(_avcodec, avcodec_find_decoder, AVCodec*(*)(enum AVCodecID))
#endif
	LOADSYM(_avcodec, avcodec_flush_buffers, void(*)(AVCodecContext*))
	LOADSYM(_avcodec, avcodec_close, int(*)(AVCodecContext*))
	LOADSYM(_avcodec, av_free_packet, void(*)(AVPacket*))
	//LOADSYM(_avcodec, av_rescale_q, int64_t (*)(int64_t, AVRational, AVRational)) // FIXME in libavutil now?

	AVInputFormat *(*dl_av_find_input_format)(const char*);
	LOADSYM(_avformat, av_find_input_format, AVInputFormat*(*)(const char*))

	void (*dl_av_register_all)(void);
	LOADSYM(_avformat, av_register_all, void(*)(void))

	// avcodec_register_all() is implicitly called by av_register_all()
	dl_av_register_all();

	// register supported extensions
	if (dl_av_find_input_format("aac"))
		_ext.append("aac|");
	if (dl_av_find_input_format("ape"))
		_ext.append("ape|apl|mac|");
	if (dl_av_find_input_format("asf"))
		_ext.append("wma|");
	if (dl_av_find_input_format("au"))
		_ext.append("au|");
	if (dl_av_find_input_format("flac"))
		_ext.append("flac|");
	if (dl_av_find_input_format("mp3"))
		_ext.append("mp2|mp3|m2a|");
	if (dl_av_find_input_format("mp4"))
		_ext.append("mp4|m4a|");
	if (dl_av_find_input_format("mpc"))
		_ext.append("mpc|");
	if (dl_av_find_input_format("ogg"))
		_ext.append("ogg|oga|");
	if (dl_av_find_input_format("rm"))
		_ext.append("rm|ra|ram|");
	if (dl_av_find_input_format("tta"))
		_ext.append("tta|");
	if (dl_av_find_input_format("voc"))
		_ext.append("voc|");
	if (dl_av_find_input_format("wav"))
		_ext.append("wav|");
	if (dl_av_find_input_format("wv"))
		_ext.append("wv|");

	_ext.replace(_ext.length()-1, ')');

	FXTRACE((10,"Ffmpeg extensions supported: %s\n", _ext.text()));
	return true;
}

bool ffmpegBackend::supports(const char* fileName) {
#if FOX_MINOR>6
	return FXPath::match(FXPath::name(fileName), _ext, NoEscape|PathName|CaseFold);
#else
	return FXPath::match(_ext, FXPath::name(fileName), FILEMATCH_FILE_NAME|FILEMATCH_NOESCAPE|FILEMATCH_CASEFOLD);
#endif
}

inline FXString ffmpegBackend::ext() const { return _ext; }

ffmpegBackend::ffmpegBackend() :
	inputBackend(name),
	_audioStream(0),
	_formatContext(0) {}

ffmpegBackend::~ffmpegBackend() {

	close();
}

FXString ffmpegBackend::getMetadata(const char* type) {

	AVDictionaryEntry* mTag=0;
	FXString info;
	for (;;) {
		mTag=dl_av_dict_get(_formatContext->metadata, type, mTag, 0);
		if (!mTag)
			break;
		if (!info.empty())
			info.append(", ");
		info.append(FXString(mTag->value).trim());
	}

	return info;
}

bool ffmpegBackend::open(const FXString& fileName) {

	close();

	AVFormatContext *fc=0;
#if LIBAVFORMAT_VERSION_MAJOR < 54
	if (dl_av_open_input_file(&fc, fileName.text(), 0, 0, 0))
#else
	if (dl_avformat_open_input(&fc, fileName.text(), 0, 0))
#endif
		return false;

	int audioStream=-1;
#if LIBAVFORMAT_VERSION_MAJOR < 54
	if (dl_av_find_stream_info(fc)<0)
#else
	if (dl_avformat_find_stream_info(fc, 0)<0)
#endif
		goto error;

	for (unsigned int i=0; i<fc->nb_streams; i++) {
		if (fc->streams[i]->codec->codec_type==AVMEDIA_TYPE_AUDIO) {
			audioStream=i;
			break;
		}
	}

	if (!openStream(fc, audioStream))
		goto error;

	_formatContext=fc;
	_audioStreamIndex=audioStream;
	_audioStream=fc->streams[audioStream];
	_decodeBufOffset=0;
	_packet.data=0;

	switch(_audioStream->codec->sample_fmt) {
	case AV_SAMPLE_FMT_U8:
#if LIBAVFORMAT_VERSION_MAJOR >= 54
	case AV_SAMPLE_FMT_U8P:
		_sampleSize = 1;
#endif
		_precision=U8;
		break;
	case AV_SAMPLE_FMT_S16:
#if LIBAVFORMAT_VERSION_MAJOR >= 54
	case AV_SAMPLE_FMT_S16P:
		_sampleSize = 2;
#endif
		_precision=S16;
		break;
	case AV_SAMPLE_FMT_S32:
#if LIBAVFORMAT_VERSION_MAJOR >= 54
	case AV_SAMPLE_FMT_S32P:
		_sampleSize = 4;
#endif
		_precision=S32;
		break;
	case AV_SAMPLE_FMT_FLT:
#if LIBAVFORMAT_VERSION_MAJOR >= 54
	case AV_SAMPLE_FMT_FLTP:
		_sampleSize = 4;
#endif
		_precision=SAMPLE_FLOAT;
		break;
	default:
		goto error;
	}

#if LIBAVCODEC_VERSION_MAJOR >= 54
	_planar = dl_av_sample_fmt_is_planar(_audioStream->codec->sample_fmt);
#endif

	_metaData.addInfo(metaData::TITLE, getMetadata("title").text());
	_metaData.addInfo(metaData::ARTIST, getMetadata("artist").text());
	_metaData.addInfo(metaData::ALBUM, getMetadata("album").text());
	_metaData.addInfo(metaData::GENRE, getMetadata("genre").text());
	_metaData.addInfo(gettext("copyright"), getMetadata("copyright").text());
	_metaData.addInfo(metaData::YEAR, getMetadata("date").text());
	_metaData.addInfo(metaData::TRACK, getMetadata("track").text());
	_metaData.addInfo(metaData::COMMENT, getMetadata("comment").text());

	time(fc->duration/1000000);

	songLoaded(fileName.text());
	return true;

error:
#if LIBAVFORMAT_VERSION_MAJOR < 54
	dl_av_close_input_file(fc);
#else
	dl_avformat_close_input(&fc);
#endif
	return false;
}

void ffmpegBackend::close() {

	if (!songLoaded())
		return;

	if (_packet.data)
		dl_av_free_packet(&_packet);

	dl_avcodec_close(_audioStream->codec);
#if LIBAVFORMAT_VERSION_MAJOR < 54
	dl_av_close_input_file(_formatContext);
#else
	dl_avformat_close_input(&_formatContext);
#endif
	_audioStream=0;
	_formatContext=0;

	songLoaded(0);
}

bool ffmpegBackend::seek(const int pos) {

	if (dl_av_seek_frame(_formatContext,
		_audioStreamIndex,
		//dl_av_rescale_q(pos, AV_TIME_BASE_Q, _audioStream->time_base),
		0,
		AVSEEK_FLAG_ANY)<0)
		return false;

	_decodeBufOffset=0;
	if (_packet.data) {
		dl_av_free_packet(&_packet);
		_packet.data=0;
	}

	dl_avcodec_flush_buffers(_formatContext->streams[_audioStreamIndex]->codec);

	return true;
}

bool ffmpegBackend::openStream(AVFormatContext* fc, const int streamIndex) {

	if ((streamIndex<0) || (streamIndex>=fc->nb_streams))
		return false;

	AVCodecContext* dec=fc->streams[streamIndex]->codec;
	AVCodec* codec=dl_avcodec_find_decoder(dec->codec_id);
	if (!codec)
		return false;

#if LIBAVCODEC_VERSION_MAJOR < 54
	if (dl_avcodec_open(dec, codec)<0)
#else
	if (dl_avcodec_open2(dec, codec, 0)<0)
#endif
		return false;

	return true;
}

/*****************************************************************/

FXIMPLEMENT(ffmpegConfig, configFrame, 0, 0)

ffmpegConfig::ffmpegConfig(FXPacker* win) :
	configFrame(win, ffmpegBackend::name, CREDITS, LINK) {

	new FXLabel(matrix(), tr("No settings available"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
}

#endif
