/*
 *  Copyright (C) 2006-2015 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "openmptBackend.h"

#ifdef HAVE_OPENMPT

#include <algorithm>
#include <string>
#include <vector>

#include "utils.h"


#ifdef HAVE_LIBZ
#  include "libs/unzip/unzip.h"
#  ifdef _WIN32
#    include "libs/unzip/iowin32.h"
#  endif
#  define EXTZ	",*.(xmz|itz|mdz|s3z)"
#endif

#define CREDITS		"libopenmpt"
#define LINK		"http://lib.openmpt.org/"

const char openmptBackend::name[]="Openmpt";

openmptConfig_t openmptBackend::_settings;

FXString openmptBackend::_ext="*.(";

/*****************************************************************/

size_t openmptBackend::fillBuffer(void* buffer, const size_t bufferSize, const unsigned int seconds) {

	size_t bufSize = (bufferSize/sizeof(float))/_settings.channels;
	return _settings.channels == 2 ?
		_module->read_interleaved_stereo(_settings.samplerate, bufSize, (float*)buffer)
		: _module->read(_settings.samplerate, bufSize, (float*)buffer);
}

/*****************************************************************/

bool openmptBackend::init() {

	std::vector<std::string> ext=openmpt::get_supported_extensions();
	for(std::vector<std::string>::iterator it=ext.begin(); it!=ext.end(); ++it) {
		_ext.append((*it).c_str()).append('|');
	}

	_ext.replace(_ext.length()-1, ')');
	_ext.append(",(mod).*");

#ifdef HAVE_LIBZ
	_ext.append(EXTZ);
#endif
	return true;
}

bool openmptBackend::supports(const char* fileName) {

	return FXPath::match(_ext, FXPath::name(fileName), FILEMATCH_FILE_NAME|FILEMATCH_NOESCAPE|FILEMATCH_CASEFOLD);
}

openmptBackend::openmptBackend() :
	inputBackend(name),
	_module(0)
{
	loadSettings();
}

openmptBackend::~openmptBackend() {

	close();
}

void openmptBackend::loadSettings() {

	_settings.samplerate=load("Frequency", 44100);
	_settings.channels=load("Channels", 2);
	_settings.resamplingMode=load("Resampling", 8);
	_settings.masterGain=load("Master gain", 0);
	_settings.stereoSeparation=load("Stereo separation", 40);
	_settings.volumeRamping=load("Volume ramping", 0);
}

void openmptBackend::saveSettings() {

	save("Frequency", _settings.samplerate);
	save("Channels", _settings.channels);
	save("Resampling", _settings.resamplingMode);
	save("Master gain", _settings.masterGain);
	save("Stereo separation", _settings.stereoSeparation);
	save("Volume ramping", _settings.volumeRamping);
}

bool openmptBackend::open(const FXString& fileName) {

	close();

	bool tmpFile=false;
	FXString fName;

#ifdef HAVE_LIBZ
	if (FXPath::match(EXTZ, FXPath::name(fileName), FILEMATCH_FILE_NAME|FILEMATCH_NOESCAPE|FILEMATCH_CASEFOLD)) {
#  ifdef _WIN32
		zlib_filefunc64_def ffunc;
		fill_win32_filefunc64(&ffunc);
#    ifdef UNICODE
		const FXnchar* tmpFileName=convertUtf(fileName);
		unzFile modZip=unzOpen2_64(tmpFileName, &ffunc);
		delete [] tmpFileName;
#    else
                unzFile modZip=unzOpen2_64(fileName.text(), &ffunc);
#    endif // UNICODE
#  else
		unzFile modZip=unzOpen64(fileName.text());
#  endif // _WIN32
		if (unzGoToFirstFile(modZip)!=UNZ_OK) {
			unzClose(modZip);
			return false;
		}
		fName=tempFile(fileName);
		FXFile modFile(fName, FXFile::Writing);
		unzOpenCurrentFile(modZip);
		FXint n;
		FXchar buffer[4096];
		do {
			n=unzReadCurrentFile(modZip, buffer, 4096);
		} while (modFile.writeBlock(buffer, n));
		unzCloseCurrentFile(modZip);
		unzClose(modZip);
		tmpFile=true;
	}
#endif // HAVE_LIBZ

	if(!tmpFile)
		fName=fileName;
		
#if defined(HAVE_MMAP) || defined(_WIN32)
	FXMemMap memMap;
	const char* data=(char*)memMap.mapFile(fName);
	if (!data) {
		delTempFile(tmpFile, fName);
		return false;
	}
	const long size=memMap.length();
#else
	FXFile f(fName);
	const long size=f.size();
	const char* data=new char[size];
	f.readBlock((void*)data, size);
	f.close();
#endif

	void* buffer=(void*)data;
	unsigned int len = size;

	try {
		_module = new openmpt::module(buffer, len);
	} catch (const openmpt::exception &e) {
		delPtr(_module);
		return false;
	}
		

#if defined(HAVE_MMAP) || defined(_WIN32)
	memMap.unmap();
#else
	delete [] data;
#endif

	delTempFile(tmpFile, fName);

	_module->set_render_param(openmpt::module::RENDER_INTERPOLATIONFILTER_LENGTH, _settings.resamplingMode);
	_module->set_render_param(openmpt::module::RENDER_MASTERGAIN_MILLIBEL, _settings.masterGain);
	_module->set_render_param(openmpt::module::RENDER_STEREOSEPARATION_PERCENT, _settings.stereoSeparation);
	_module->set_render_param(openmpt::module::RENDER_VOLUMERAMPING_STRENGTH, _settings.volumeRamping);

	std::vector<std::string> metadata=_module->get_metadata_keys();
	for(std::vector<std::string>::iterator it=metadata.begin(); it!=metadata.end(); ++it) {
		FXTRACE((10, "%s\n", (*it).c_str()));
		if ((*it).compare("title") == 0)
			_metaData.addInfo(metaData::TITLE, _module->get_metadata("title").c_str());
		else
		if ((*it).compare("artist") == 0)
			_metaData.addInfo(metaData::ARTIST, _module->get_metadata("artist").c_str());
		else
		if ((*it).compare("message") == 0)
			_metaData.addInfo(metaData::COMMENT, _module->get_metadata("message").c_str());
	}

	time(_module->get_duration_seconds());

	songLoaded(fileName.text());
	return true;
}

void openmptBackend::close() {

	if (_module) {
		delPtr(_module);
	}

	songLoaded(0);
}

bool openmptBackend::rewind() {

	if (_module) {
		_module->set_position_order_row(0, 0);
		return true;
	}
	return false;
}

/*****************************************************************/

#define SETTINGS	openmptBackend::_settings

FXDEFMAP(openmptConfig) openmptConfigMap[] = {
	// Message_Type		ID				Message_Handler
	FXMAPFUNC(SEL_COMMAND,	openmptConfig::ID_FREQUENCY,	openmptConfig::onCmdFrequency),
	FXMAPFUNC(SEL_COMMAND,	openmptConfig::ID_CHANNELS,		openmptConfig::onCmdChannels),
	FXMAPFUNC(SEL_COMMAND,	openmptConfig::ID_RESAMPLING,	openmptConfig::onCmdResampling),
};

FXIMPLEMENT(openmptConfig, configFrame, openmptConfigMap, ARRAYNUMBER(openmptConfigMap))

openmptConfig::openmptConfig(FXPacker* win) :
	configFrame(win, openmptBackend::name, CREDITS, LINK) {

	FXint val;

	new FXLabel(matrix(), tr("Frequency"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *freqBox=new FXListBox(matrix(), this, openmptConfig::ID_FREQUENCY, LAYOUT_FILL|FRAME_NORMAL);
	freqBox->fillItems("11025\n22050\n44100\n48000");
	freqBox->setNumVisible(4);
	switch (SETTINGS.samplerate) {
	case 11025:
		val=0;
		break;
	case 22050:
		val=1;
		break;
	default:
	case 44100:
		val=2;
		break;
	case 48000:
		val=3;
		break;
	}
	freqBox->setCurrentItem(val);

	new FXLabel(matrix(), tr("Channels"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *chanBox=new FXListBox(matrix(), this, openmptConfig::ID_CHANNELS, LAYOUT_FILL|FRAME_NORMAL);
	chanBox->fillItems("Mono\nStereo");
	chanBox->setNumVisible(2);
	chanBox->setCurrentItem(SETTINGS.channels-1);

	new FXLabel(matrix(), tr("Resampling"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *resBox=new FXListBox(matrix(), this, openmptConfig::ID_RESAMPLING, LAYOUT_FILL|FRAME_NORMAL);
	resBox->fillItems("Zero order hold\nLinear\nCubic Spline\nSinc 8 taps");
	resBox->setNumVisible(4);
	switch (SETTINGS.resamplingMode) {
	case 1:
		val=0;
		break;
	case 2:
		val=1;
		break;
	default:
	case 4:
		val=2;
		break;
	case 8:
		val=3;
		break;
	}
	resBox->setCurrentItem(val);

	//
	FXHorizontalFrame *hf=new FXHorizontalFrame(extraBottom(), 0, 0,0,0,0, 0,0,0,0, 0,0);

	_masterGain.connect(SETTINGS.masterGain);
	_stereoSeparation.connect(SETTINGS.stereoSeparation);
	_volumeRamping.connect(SETTINGS.volumeRamping);

	FXKnob		*knob;
	FXTextField	*tf;
	FXMatrix	*mat;

	FXVerticalFrame *vert=new FXVerticalFrame(hf, FRAME_GROOVE);
	mat=new FXMatrix(vert, 3, MATRIX_BY_ROWS|LAYOUT_FILL|PACK_UNIFORM_WIDTH);

	new FXLabel(mat, tr("Master Gain"));
	knob=new FXKnob(mat, &_masterGain, FXDataTarget::ID_VALUE, LAYOUT_CENTER_Y|KNOB_NORMAL);
	knob->setRange(0, 100);
	knob->setTickDelta(10);
	tf=new FXTextField(mat, 3, &_masterGain, FXDataTarget::ID_VALUE, TEXTFIELD_READONLY|TEXTFIELD_NORMAL|LAYOUT_CENTER_X|JUSTIFY_RIGHT);
	tf->disable();

	new FXLabel(mat, tr("Stereo Separation"));
	knob=new FXKnob(mat, &_stereoSeparation, FXDataTarget::ID_VALUE, LAYOUT_CENTER_Y|KNOB_NORMAL);
	knob->setRange(0, 200);
	knob->setTickDelta(20);
	tf=new FXTextField(mat, 3, &_stereoSeparation, FXDataTarget::ID_VALUE, TEXTFIELD_READONLY|TEXTFIELD_NORMAL|LAYOUT_CENTER_X|JUSTIFY_RIGHT);
	tf->disable();

	new FXLabel(mat, tr("Volume Ramping"));
	knob=new FXKnob(mat, &_volumeRamping, FXDataTarget::ID_VALUE, LAYOUT_CENTER_Y|KNOB_NORMAL);
	knob->setRange(-1, 10);
	knob->setTickDelta(1);
	tf=new FXTextField(mat, 3, &_volumeRamping, FXDataTarget::ID_VALUE, TEXTFIELD_READONLY|TEXTFIELD_NORMAL|LAYOUT_CENTER_X|JUSTIFY_RIGHT);
	tf->disable();
}

long openmptConfig::onCmdFrequency(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	switch (val) {
	case 0:
		SETTINGS.samplerate=11025;
		break;
	case 1:
		SETTINGS.samplerate=22050;
		break;
	case 2:
		SETTINGS.samplerate=44100;
		break;
	case 3:
		SETTINGS.samplerate=48000;
		break;
	}
	return 1;
}

long openmptConfig::onCmdChannels(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	SETTINGS.channels=val+1;
	return 1;
}

long openmptConfig::onCmdResampling(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	switch (val) {
	case 0:
		SETTINGS.resamplingMode=1;
		break;
	case 1:
		SETTINGS.resamplingMode=2;
		break;
	case 2:
		SETTINGS.resamplingMode=4;
		break;
	case 3:
		SETTINGS.resamplingMode=8;
		break;
	}
	return 1;
}

#endif
