/*
 *  Copyright (C) 2006-2015 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "sidBackend.h"

#ifdef HAVE_SIDPLAYFP

#include <sidplayfp/SidInfo.h>

#include <sidplayfp/builders/residfp.h>
#include <sidplayfp/builders/resid.h>
#include <sidplayfp/builders/hardsid.h>

#include <FX88591Codec.h>
#include "utils.h"

// created by reswrap from file sid.gif
extern const unsigned char iconSid[126]={
  0x47,0x49,0x46,0x38,0x39,0x61,0x10,0x00,0x10,0x00,0xf2,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x86,0x00,0x00,0xff,0x86,0x00,0x00,0xff,0x00,0x00,0x86,0x82,0x86,0xcc,
  0xcc,0xcc,0x00,0x00,0x00,0x21,0xf9,0x04,0x01,0x00,0x00,0x06,0x00,0x2c,0x00,0x00,
  0x00,0x00,0x10,0x00,0x10,0x00,0x00,0x03,0x43,0x68,0xba,0xdc,0x6e,0x22,0x3e,0x16,
  0x6b,0x99,0xb0,0xc6,0xfb,0x44,0xf0,0x45,0xa8,0x09,0xcc,0x17,0x84,0x4a,0xa0,0x9e,
  0x9c,0x89,0xa6,0x00,0xbb,0xac,0xdc,0x12,0xd6,0x06,0xcd,0x10,0xbc,0xb3,0x32,0x83,
  0xe0,0xa0,0xa1,0x8a,0x2d,0x06,0x00,0xe4,0x90,0x11,0x6b,0x06,0x0c,0xb7,0xd7,0xc2,
  0x09,0xc0,0x3d,0x0a,0x80,0xac,0x75,0x72,0xc3,0x78,0x17,0x09,0x00,0x3b
  };

//sid|dat
#define EXT	"*.(sid|mus|prg|p00)"

// HVSC path to STIL.
#define HVSC_STIL	"/DOCUMENTS/STIL.txt"

// HVSC path to BUGlist.
#define HVSC_BUGLIST	"/DOCUMENTS/BUGlist.txt"

#define CREDITS		"Sidplayfp\nCopyright \302\251 Simon White\nAntti Lankila\nLeandro Nini"
#define LINK		"http://sourceforge.net/projects/sidplay-residfp/"

const char sidBackend::name[]="Sidplayfp";


const char sidBackend::_engines[3][8] = {"reSID", "hardSID", "reSIDfp"};

const int sidAddresses[5] = { 0, 0xd420, 0xd500, 0xde00, 0xdf00 };

sidConfig_t sidBackend::_settings;

/*****************************************************************/

size_t sidBackend::fillBuffer(void* buffer, const size_t bufferSize, const unsigned int seconds) {

	if (_length && (seconds>=_length))
		return 0;

	return _sidplayfp->play((short*)buffer, bufferSize/sizeof(short))*2;
}

/*****************************************************************/

const char* getModelString(SidTuneInfo::model_t model) {
	switch (model) {
		default:
		case SidTuneInfo::SIDMODEL_UNKNOWN:
			return "UNKNOWN";
		case SidTuneInfo::SIDMODEL_6581:
			return "6581";
		case SidTuneInfo::SIDMODEL_8580:
			return "8580";
		case SidTuneInfo::SIDMODEL_ANY:
			return "ANY";
	}
}

const char* getClockString(SidTuneInfo::clock_t clock) {
	switch (clock) {
		default:
		case SidTuneInfo::CLOCK_UNKNOWN:
			return "UNKNOWN";
		case SidTuneInfo::CLOCK_PAL:
			return "PAL";
		case SidTuneInfo::CLOCK_NTSC:
			return "NTSC";
		case SidTuneInfo::CLOCK_ANY:
			return "ANY";
	}
}

/*****************************************************************/

bool sidBackend::supports(const char* fileName) {

#if FOX_MINOR>6
	return FXPath::match(FXPath::name(fileName), EXT, NoEscape|PathName|CaseFold);
#else
	return FXPath::match(FXString(EXT), FXPath::name(fileName), FILEMATCH_FILE_NAME|FILEMATCH_NOESCAPE|FILEMATCH_CASEFOLD);
#endif
}

inline FXString sidBackend::ext() const { return FXString(EXT); }

sidBackend::sidBackend() :
	inputBackend(name, iconSid),
	_sidplayfp(0),
	_stil(0),
	_db(0),
	_length(0),
	_tune(0) {

	loadSettings();

	openHvsc(_settings.hvscPath);
}

sidBackend::~sidBackend() {

	close();

	delete _db;
	delete _stil;
}

void sidBackend::loadSettings() {

	_settings.samplerate=load("Frequency", 44100);
	_settings.channels=load("Channels", 1);
	_settings.samplingMethod=(SidConfig::sampling_method_t)load("Sampling method", SidConfig::INTERPOLATE);
	_settings.fastSampling=load("Fast sampling", FALSE);
	_settings.bias=load("DAC Bias", 0);
	_settings.filter6581Curve=load("Filter 6581 Curve", 500);
	_settings.filter8580Curve=load("Filter 8580 Curve", 12500);
	_settings.c64Model=(SidConfig::c64_model_t)load("C64 Model", SidConfig::PAL);
	_settings.sidModel=(SidConfig::sid_model_t)load("SID model", SidConfig::MOS6581);
	_settings.forceC64Model=load("Force C64 Model", FALSE);
	_settings.forceSidModel=load("Force SID Model", FALSE);
	_settings.engine=load("Engine", _engines[0]);
	_settings.filter=load("Filter", TRUE);
	_settings.hvscPath=load("HVSC", FXString::null);
	_settings.secondSidAddress=load("Second SID address", 0);
#ifdef ENABLE_3SID
	_settings.thirdSidAddress=load("Third SID address", 0);
#endif
	_settings.kernalPath=load("Kernal Rom", FXString::null);
	_settings.basicPath=load("BASIC Rom", FXString::null);
	_settings.chargenPath=load("Chargen Rom", FXString::null);
}

void sidBackend::saveSettings() {

	if (compare(_settings.hvscPath, load("HVSC", FXString::null))) {
		FXTRACE((10, "Reloading HVSC from %s\n", _settings.hvscPath.text()));
		openHvsc(_settings.hvscPath);
	}

	save("Frequency", _settings.samplerate);
	save("Channels", _settings.channels);
	save("Sampling method", _settings.samplingMethod);
	save("Fast sampling", _settings.fastSampling);
	save("DAC Bias", _settings.bias);
	save("Filter 6581 Curve", _settings.filter6581Curve);
	save("Filter 8580 Curve", _settings.filter8580Curve);
	save("C64 Model", _settings.c64Model);
	save("SID model", _settings.sidModel);
	save("Force C64 Model", _settings.forceC64Model);
	save("Force SID Model", _settings.forceSidModel);
	save("Engine", _settings.engine);
	save("Filter", _settings.filter);
	save("HVSC", _settings.hvscPath);
	save("Second SID address", _settings.secondSidAddress);
#ifdef ENABLE_3SID
	save("Third SID address", _settings.thirdSidAddress);
#endif
	save("Kernal Rom", _settings.kernalPath);
	save("BASIC Rom", _settings.basicPath);
	save("Chargen Rom", _settings.chargenPath);
}

const unsigned char* sidBackend::loadRom(const FXString& romPath) {
	FXFile f;
	if (!f.open(romPath))
		return 0;
	const long size=f.size();
	const unsigned char* data=new unsigned char[size];
	f.readBlock((void*)data, size);
	f.close();
	return data;
}

bool sidBackend::open(const FXString& fileName) {

	close();

	_sidplayfp=new sidplayfp;

	{
		const unsigned char* kernal=loadRom(_settings.kernalPath.text());
		const unsigned char* basic=loadRom(_settings.basicPath.text());
		const unsigned char* chargen=loadRom(_settings.chargenPath.text());
		_sidplayfp->setRoms(kernal, basic, chargen);
		delete [] kernal;
		delete [] basic;
		delete [] chargen;
	}
	sidbuilder *emuSid=0;

	if (!comparecase(_settings.engine, _engines[2])) {
		ReSIDfpBuilder *tmpResid=new ReSIDfpBuilder("FAP reSIDfp");
		tmpResid->create(_sidplayfp->info().maxsids());

		tmpResid->filter(_settings.filter);
		tmpResid->filter6581Curve((double)_settings.filter6581Curve/1000.);
		tmpResid->filter8580Curve((double)_settings.filter8580Curve);

		emuSid=(sidbuilder*)tmpResid;
	}

	if (!comparecase(_settings.engine, _engines[0])) {
		ReSIDBuilder *tmpResid=new ReSIDBuilder("FAP reSID");
		tmpResid->create(_sidplayfp->info().maxsids());
		tmpResid->filter(_settings.filter);

		tmpResid->bias((double)_settings.bias/1000.0);
		emuSid=(sidbuilder*)tmpResid;
	}

	if (!comparecase(_settings.engine, _engines[1])) {
		HardSIDBuilder *tmpHardsid=new HardSIDBuilder("FAP hardSID");
		tmpHardsid->create(_sidplayfp->info().maxsids());
		tmpHardsid->filter(_settings.filter);
		emuSid=(sidbuilder*)tmpHardsid;
	}

	if (!emuSid) {
		delPtr(_sidplayfp);
		return false;
	}

	SidConfig cfg;
	cfg.defaultC64Model=_settings.c64Model;
	cfg.forceC64Model=_settings.forceC64Model;
	cfg.defaultSidModel=_settings.sidModel;
	cfg.forceSidModel=_settings.forceSidModel;
	cfg.playback=(_settings.channels==2)?SidConfig::STEREO:SidConfig::MONO;
	cfg.frequency=_settings.samplerate;
	cfg.secondSidAddress=_settings.secondSidAddress;
#ifdef ENABLE_3SID
	cfg.thirdSidAddress=_settings.thirdSidAddress;
#endif
	cfg.sidEmulation=emuSid;
	cfg.samplingMethod=_settings.samplingMethod;
	cfg.fastSampling=_settings.fastSampling;
	_sidplayfp->config(cfg);

	_tune=new SidTune(fileName.text());
	_tune->createMD5(_md5);
	FXTRACE((10, "Tune md5: %s\n", _md5));

	loadTune(0);

	const FX88591Codec codec;
	const SidTuneInfo* tuneInfo=_tune->getInfo();

	/*
	* SID tunes have three info strings (title, artis, released)
	* p00 files have only title
	* prg files have none
	*/
	switch (tuneInfo->numberOfInfoStrings()) {
	case 3:
		_metaData.addInfo(gettext("released"), codec.mb2utf(tuneInfo->infoString(2)).text());
	case 2:
		_metaData.addInfo(metaData::ARTIST, codec.mb2utf(tuneInfo->infoString(1)).text());
	case 1:
		_metaData.addInfo(metaData::TITLE, codec.mb2utf(tuneInfo->infoString(0)).text());;
	}

	/*
	* MUS files have only comments
	*/
	const unsigned int n = tuneInfo->numberOfCommentStrings();
	if (n) {
		FXString info;
		for (unsigned int i = 0; i < n; i++) {
			if (!info.empty())
				info.append('\n');
			info.append(tuneInfo->commentString(i));
		}

		if (!info.empty()) {
			_metaData.addInfo(metaData::COMMENT, info.text());
		}
	}

	if (_stil) {
		FXTRACE((10, "Retrieving STIL info\n"));
		FXString comment=FXString(_stil->getAbsGlobalComment(fileName.text()));
		if (!comment.empty())
			comment.append('\n');

		comment.append(FXString(_stil->getAbsEntry(fileName.text())));

		FXString bug=FXString(_stil->getAbsBug(fileName.text()));
		if (!bug.empty()) {
			comment.append('\n');
			comment.append(bug);
		}
		_metaData.addInfo(metaData::COMMENT, codec.mb2utf(comment.text()).text());
	}

	_metaData.addInfo(gettext("speed"), _sidplayfp->info().speedString());
	_metaData.addInfo(gettext("file format"), tuneInfo->formatString());
	_metaData.addInfo(gettext("song clock"), getClockString(tuneInfo->clockSpeed()));
#ifdef ENABLE_3SID
	_metaData.addInfo(gettext("SID model"), getModelString(tuneInfo->sidModel(0)));
	if (tuneInfo->sidChips() > 1) {
		_metaData.addInfo(gettext("2nd SID model"), getModelString(tuneInfo->sidModel(1)));
		_metaData.addInfo(gettext("2nd SID address"), FXStringFormat("$%4x", tuneInfo->sidChipBase(1)).text());
		if (tuneInfo->sidChips() > 2) {
			_metaData.addInfo(gettext("3rd SID model"), getModelString(tuneInfo->sidModel(2)));
			_metaData.addInfo(gettext("3rd SID address"), FXStringFormat("$%4x", tuneInfo->sidChipBase(2)).text());
		}
	}
#else
        _metaData.addInfo(gettext("SID model"), getModelString(tuneInfo->sidModel1()));
	if (tuneInfo->isStereo()) {
		_metaData.addInfo(gettext("2nd SID model"), getModelString(tuneInfo->sidModel2()));
		_metaData.addInfo(gettext("2nd SID address"), FXStringFormat("$%4x", tuneInfo->sidChipBase2()).text());
	}
#endif
	songLoaded(fileName.text());
	return true;
}

void sidBackend::close() {

	if (_sidplayfp) {
		const sidbuilder *emuSid=_sidplayfp->config().sidEmulation;
		delete emuSid;
		delPtr(_sidplayfp);
	}

        delPtr(_tune);
    
	songLoaded(0);
}

bool sidBackend::rewind() {

	if (_sidplayfp) {
		_sidplayfp->stop();
		return true;
	}
	return false;
}

bool sidBackend::subtune(const unsigned int i) {

	if (_tune && (i<=_tune->getInfo()->songs())) {
		loadTune(i);
		return true;
	}
	return false;
}

void sidBackend::loadTune(const int num) {

	_tune->selectSong(num);
	_sidplayfp->load(_tune);
	_length=_db?_db->length(_md5, _tune->getInfo()->currentSong()):0;
	time(_length);
}

void sidBackend::openHvsc(const FXString& hvscPath) {

	if (hvscPath.empty())
		return;

	if (!_stil)
		_stil=new STIL(HVSC_STIL, HVSC_BUGLIST);

	if (!_stil->setBaseDir(hvscPath.text())) {
		fxmessage("%s\n", _stil->getErrorStr());
		delPtr(_stil);
	}

	if (!_db)
		_db=new SidDatabase();

	if (!_db->open(FXStringFormat("%s%cDOCUMENTS%cSonglengths.txt", hvscPath.text(), PATHSEP, PATHSEP).text())) {
		fxmessage("%s\n", _db->error());
		delPtr(_db);
	}
}

/*****************************************************************/

#include "iconFactory.h"

#define SETTINGS	sidBackend::_settings

FXDEFMAP(sidConfig) sidConfigMap[] = {
	// Message_Type		ID				Message_Handler
	FXMAPFUNC(SEL_COMMAND,	sidConfig::ID_FREQUENCY,	sidConfig::onCmdFrequency),
	FXMAPFUNC(SEL_COMMAND,	sidConfig::ID_CHANNELS,		sidConfig::onCmdChannels),
	FXMAPFUNC(SEL_COMMAND,	sidConfig::ID_SAMPLING,		sidConfig::onCmdSampling),
	FXMAPFUNC(SEL_COMMAND,	sidConfig::ID_ENGINE,		sidConfig::onCmdEngine),
	FXMAPFUNC(SEL_COMMAND,	sidConfig::ID_CLOCK,		sidConfig::onCmdClock),
	FXMAPFUNC(SEL_COMMAND,	sidConfig::ID_MODEL,		sidConfig::onCmdModel),
	FXMAPFUNC(SEL_COMMAND,  sidConfig::ID_ADDRESS,		sidConfig::onCmdAddress),
	FXMAPFUNC(SEL_COMMAND,	sidConfig::ID_HVSC,		sidConfig::onCmdHvsc),
	FXMAPFUNC(SEL_COMMAND,	sidConfig::ID_KERNAL,		sidConfig::onCmdRom),
	FXMAPFUNC(SEL_COMMAND,	sidConfig::ID_BASIC,		sidConfig::onCmdRom),
	FXMAPFUNC(SEL_COMMAND,	sidConfig::ID_CHARGEN,		sidConfig::onCmdRom),
};

FXIMPLEMENT(sidConfig, configFrame, sidConfigMap, ARRAYNUMBER(sidConfigMap))

sidConfig::sidConfig(FXPacker* win) :
	configFrame(win, sidBackend::name, CREDITS, LINK) {

	FXint val;

	_fastSampling.connect(SETTINGS.fastSampling);
	_forceC64Model.connect(SETTINGS.forceC64Model);
        _forceSidModel.connect(SETTINGS.forceSidModel);
	_filter.connect(SETTINGS.filter);
	_hvscDir.connect(SETTINGS.hvscPath);
	_kernalDir.connect(SETTINGS.kernalPath);
	_basicDir.connect(SETTINGS.basicPath);
	_chargenDir.connect(SETTINGS.chargenPath);

	new FXLabel(matrix(), tr("Frequency"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *freqBox=new FXListBox(matrix(), this, sidConfig::ID_FREQUENCY, LAYOUT_FILL|FRAME_NORMAL);
	freqBox->fillItems("11025\n22050\n44100\n48000");
	freqBox->setNumVisible(4);
	switch (SETTINGS.samplerate) {
	case 11025:
		val=0;
		break;
	case 22050:
		val=1;
		break;
	default:
	case 44100:
		val=2;
		break;
	case 48000:
		val=3;
		break;
	}
	freqBox->setCurrentItem(val);

	new FXLabel(matrix(), tr("Channels"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *chanBox=new FXListBox(matrix(), this, sidConfig::ID_CHANNELS, LAYOUT_FILL|FRAME_NORMAL);
	chanBox->fillItems("Mono\nStereo");
	chanBox->setNumVisible(2);
	chanBox->setCurrentItem(SETTINGS.channels-1);

	new FXLabel(matrix(), tr("Engine"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *engBox=new FXListBox(matrix(), this, sidConfig::ID_ENGINE, LAYOUT_FILL|FRAME_NORMAL);
	engBox->appendItem(sidBackend::_engines[0]);
	engBox->appendItem(sidBackend::_engines[1]);
	engBox->appendItem(sidBackend::_engines[2]);

	const FXint numItems=engBox->getNumItems();
	engBox->setNumVisible(numItems);
	const FXint curItem=engBox->findItem(SETTINGS.engine);
	if (curItem>=0)
		engBox->setCurrentItem(curItem);

	new FXLabel(matrix(), tr("Sampling method"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *resBox=new FXListBox(matrix(), this, sidConfig::ID_SAMPLING, LAYOUT_FILL|FRAME_NORMAL);
	resBox->fillItems("Interpolate\nResample Interpolate");
	resBox->setNumVisible(2);
	switch (SETTINGS.samplingMethod) {
	default:
	case SidConfig::INTERPOLATE:
		val=0;
		break;
	case SidConfig::RESAMPLE_INTERPOLATE:
		val=1;
		break;
	}
	resBox->setCurrentItem(val);

	new FXLabel(matrix(), tr("C64 model"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *clockBox=new FXListBox(matrix(), this, sidConfig::ID_CLOCK, LAYOUT_FILL|FRAME_NORMAL);
	clockBox->fillItems("PAL\nNTSC\nOLD NTSC\nDrean");
	clockBox->setNumVisible(4);
	switch (SETTINGS.c64Model) {
	default:
	case SidConfig::PAL:
		val=0;
		break;
	case SidConfig::NTSC:
		val=1;
		break;
	case SidConfig::OLD_NTSC:
		val=2;
		break;
	case SidConfig::DREAN:
		val=3;
		break;
	}
	clockBox->setCurrentItem(val);

	new FXLabel(matrix(), tr("Force C64 model"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	new FXCheckButton(matrix(), FXString::null, &_forceC64Model, FXDataTarget::ID_VALUE);

	new FXLabel(matrix(), tr("SID Model"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *modelBox=new FXListBox(matrix(), this, sidConfig::ID_MODEL, LAYOUT_FILL|FRAME_NORMAL);
	modelBox->fillItems("MOS 6581\nMOS 8580");
	modelBox->setNumVisible(2);
	switch (SETTINGS.sidModel) {
	default:
	case SidConfig::MOS6581:
		val=0;
		break;
	case SidConfig::MOS8580:
		val=1;
		break;
	}
	modelBox->setCurrentItem(val);

	new FXLabel(matrix(), tr("Force SID model"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	new FXCheckButton(matrix(), FXString::null, &_forceSidModel, FXDataTarget::ID_VALUE);

	new FXLabel(matrix(), tr("Second SID address"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *sidAddress=new FXListBox(matrix(), this, sidConfig::ID_ADDRESS, LAYOUT_FILL|FRAME_NORMAL);
	sidAddress->fillItems("(none)\n$D420\n$D500\n$DE00\n$DF00");
	sidAddress->setNumVisible(5);
	val = 5;
	while (val--) {
		if (SETTINGS.secondSidAddress == sidAddresses[val])
			break;
        }
	sidAddress->setCurrentItem(val);

#ifdef ENABLE_3SID
	new FXLabel(matrix(), tr("Third SID address"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	sidAddress=new FXListBox(matrix(), this, sidConfig::ID_ADDRESS, LAYOUT_FILL|FRAME_NORMAL);
	sidAddress->fillItems("(none)\n$D420\n$D500\n$DE00\n$DF00");
	sidAddress->setNumVisible(5);
	val = 5;
	while (val--) {
		if (SETTINGS.thirdSidAddress == sidAddresses[val])
			break;
        }
	sidAddress->setCurrentItem(val);
#endif

	new FXVerticalSeparator(extraLeft());

	FXVerticalFrame *vert=new FXVerticalFrame(extraLeft());
	new FXCheckButton(vert, tr("Fast sampling\tFaster but inaccurate sampling"), &_fastSampling, FXDataTarget::ID_VALUE);
	new FXCheckButton(vert, tr("Filter\tEmulate SID filter"), &_filter, FXDataTarget::ID_VALUE);

	_bias.connect(SETTINGS.bias);
	_filter6581Curve.connect(SETTINGS.filter6581Curve);
	_filter8580Curve.connect(SETTINGS.filter8580Curve);

	_biasFrame=new FXVerticalFrame(vert, FRAME_GROOVE|LAYOUT_FILL_X);
	new FXLabel(_biasFrame, tr("DAC Bias for reSID"));
	FXMatrix *mat=new FXMatrix(_biasFrame, 2, MATRIX_BY_COLUMNS|LAYOUT_FILL|PACK_UNIFORM_WIDTH);
	FXKnob* knob=new FXKnob(mat, &_bias, FXDataTarget::ID_VALUE, LAYOUT_CENTER_Y|KNOB_NORMAL);
	knob->setRange(-500, 500);
	knob->setTickDelta(100);
	FXTextField *tf=new FXTextField(mat, 3, &_bias, FXDataTarget::ID_VALUE,
		TEXTFIELD_READONLY|TEXTFIELD_NORMAL|LAYOUT_CENTER_Y|JUSTIFY_RIGHT);
	tf->disable();

	_filterCurveFrame=new FXVerticalFrame(vert, FRAME_GROOVE|LAYOUT_FILL_X);
	new FXLabel(_filterCurveFrame, tr("Filter curves for reSIDfp"));
	mat=new FXMatrix(_filterCurveFrame, 2, MATRIX_BY_COLUMNS|LAYOUT_FILL|PACK_UNIFORM_WIDTH);
	new FXLabel(mat, "6581");
	new FXLabel(mat, "8580");
	knob=new FXKnob(mat, &_filter6581Curve, FXDataTarget::ID_VALUE, LAYOUT_CENTER_Y|KNOB_NORMAL);
	knob->setRange(0, 1000);
	knob->setTickDelta(100);
	knob=new FXKnob(mat, &_filter8580Curve, FXDataTarget::ID_VALUE, LAYOUT_CENTER_Y|KNOB_NORMAL);
	knob->setRange(8000, 16000);
	knob->setTickDelta(500);
	tf=new FXTextField(mat, 4, &_filter6581Curve, FXDataTarget::ID_VALUE,
		TEXTFIELD_READONLY|TEXTFIELD_NORMAL|LAYOUT_CENTER_Y|JUSTIFY_RIGHT);
	tf->disable();
	tf=new FXTextField(mat, 5, &_filter8580Curve, FXDataTarget::ID_VALUE,
		TEXTFIELD_READONLY|TEXTFIELD_NORMAL|LAYOUT_CENTER_Y|JUSTIFY_RIGHT);
	tf->disable();

	FXMatrix *frame=new FXMatrix(extraBottom(), 3, MATRIX_BY_COLUMNS);

	new FXLabel(frame, tr("HVSC path:"));
	new FXTextField(frame, 40, &_hvscDir, FXDataTarget::ID_VALUE, TEXTFIELD_READONLY|TEXTFIELD_NORMAL);
	new FXButton(frame, tr("&Browse\tSelect HVSC directory"), GET_ICON(icon_documentopen), this, sidConfig::ID_HVSC, FRAME_RAISED|FRAME_THICK|JUSTIFY_LEFT|ICON_BEFORE_TEXT);

	new FXLabel(frame, tr("Kernal Rom:"));
	new FXTextField(frame, 40, &_kernalDir, FXDataTarget::ID_VALUE, TEXTFIELD_READONLY|TEXTFIELD_NORMAL);
	new FXButton(frame, tr("&Browse\tSelect Kernal Rom file"), GET_ICON(icon_documentopen), this, sidConfig::ID_KERNAL, FRAME_RAISED|FRAME_THICK|JUSTIFY_LEFT|ICON_BEFORE_TEXT);

	new FXLabel(frame, tr("BASIC Rom:"));
	new FXTextField(frame, 40, &_basicDir, FXDataTarget::ID_VALUE, TEXTFIELD_READONLY|TEXTFIELD_NORMAL);
	new FXButton(frame, tr("&Browse\tSelect BASIC Rom file"), GET_ICON(icon_documentopen), this, sidConfig::ID_BASIC, FRAME_RAISED|FRAME_THICK|JUSTIFY_LEFT|ICON_BEFORE_TEXT);

	new FXLabel(frame, tr("Chargen Rom:"));
	new FXTextField(frame, 40, &_chargenDir, FXDataTarget::ID_VALUE, TEXTFIELD_READONLY|TEXTFIELD_NORMAL);
	new FXButton(frame, tr("&Browse\tSelect Chargen Rom file"), GET_ICON(icon_documentopen), this, sidConfig::ID_CHARGEN, FRAME_RAISED|FRAME_THICK|JUSTIFY_LEFT|ICON_BEFORE_TEXT);
}

long sidConfig::onCmdFrequency(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	switch (val) {
	case 0:
		SETTINGS.samplerate=11025;
		break;
	case 1:
		SETTINGS.samplerate=22050;
		break;
	case 2:
		SETTINGS.samplerate=44100;
		break;
	case 3:
		SETTINGS.samplerate=48000;
		break;
	}
	return 1;
}

long sidConfig::onCmdChannels(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	SETTINGS.channels=val+1;
	return 1;
}

long sidConfig::onCmdSampling(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	switch (val) {
	case 0:
		SETTINGS.samplingMethod=SidConfig::INTERPOLATE;
		break;
	case 1:
		SETTINGS.samplingMethod=SidConfig::RESAMPLE_INTERPOLATE;
		break;
	}
	return 1;
}

long sidConfig::onCmdEngine(FXObject* obj, FXSelector, void* ptr) {

	FXListBox *engBox=static_cast<FXListBox*>(obj);
	const FXint val=(FXint)(FXival)ptr;
	SETTINGS.engine=engBox->getItemText(val);

	/*if (!comparecase(SETTINGS.engine, sidBackend::_engines[0])) {
		_biasFrame->show();
		_filterCurveFrame->hide();
	} else if (!comparecase(SETTINGS.engine, sidBackend::_engines[2])) {
		_biasFrame->hide();
		_filterCurveFrame->show();
	} else {
		_biasFrame->hide();
		_filterCurveFrame->hide();
	}*/

	return 1;
}

long sidConfig::onCmdClock(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	switch (val) {
	case 0:
		SETTINGS.c64Model=SidConfig::PAL;
		break;
	case 1:
		SETTINGS.c64Model=SidConfig::NTSC;
		break;
	case 2:
		SETTINGS.c64Model=SidConfig::OLD_NTSC;
		break;
	case 3:
		SETTINGS.c64Model=SidConfig::DREAN;
		break;
	}
	return 1;
}

long sidConfig::onCmdModel(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	switch (val) {
	case 0:
		SETTINGS.sidModel=SidConfig::MOS6581;
		break;
	case 1:
		SETTINGS.sidModel=SidConfig::MOS8580;
		break;
	}
	return 1;
}

long sidConfig::onCmdAddress(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	SETTINGS.secondSidAddress = sidAddresses[val];
	return 1;
}

long sidConfig::onCmdHvsc(FXObject*, FXSelector, void*) {

	FXDirDialog dirDialog(this, tr("Select HVSC directory"));
	dirDialog.setDirectory(SETTINGS.hvscPath);
	if (dirDialog.execute())
		SETTINGS.hvscPath=dirDialog.getDirectory();

	return 1;
}

long sidConfig::onCmdRom(FXObject*, FXSelector sel, void*) {

	const char* text;
	FXString* romPath;
	switch (FXSELID(sel)) {
	case ID_KERNAL:
		text = "Select Kernal Rom file";
		romPath = &SETTINGS.kernalPath;
		break;
	case ID_BASIC:
		text = "Select BASIC Rom file";
		romPath = &SETTINGS.basicPath;
		break;
	case ID_CHARGEN:
		text = "Select Chargen Rom file";
		romPath = &SETTINGS.chargenPath;
		break;
	default:
		return 0;
	}
	FXFileDialog fileDialog(this, tr(text));
	fileDialog.setFilename(*romPath);
	if (fileDialog.execute())
		*romPath=fileDialog.getFilename();

	return 1;
}

#endif
