/*
 *  Copyright (C) 2006-2015 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef SID_H
#define SID_H

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#ifdef HAVE_SIDPLAYFP

#include <sidplayfp/sidplayfp.h>
#include <sidplayfp/SidDatabase.h>
#include <sidplayfp/SidTune.h>
#include <sidplayfp/SidConfig.h>
#include <sidplayfp/SidTuneInfo.h>

#include <fx.h>
#include "inputBackend.h"

#ifdef HAVE_STILVIEW
#  include <stilview/stil.h>
#else
#  include "stilview/stil.h"
#endif

#if LIBSIDPLAYFP_VERSION_MAJ == 1 && LIBSIDPLAYFP_VERSION_MIN > 7
#  define ENABLE_3SID
#endif


/*****************************************************************/

typedef struct {
	FXString	engine;
	FXint		samplerate;
	FXint		channels;
	SidConfig::sampling_method_t	samplingMethod;
	FXbool		fastSampling;
	FXint		bias;
	FXint		filter6581Curve;
	FXint		filter8580Curve;
	SidConfig::c64_model_t	c64Model;
	FXbool		forceC64Model;
	SidConfig::sid_model_t	sidModel;
	FXbool		forceSidModel;
	FXbool		filter;
	FXString	hvscPath;
        FXint           secondSidAddress;
#ifdef ENABLE_3SID
	FXint           thirdSidAddress;
#endif
	FXString	kernalPath;
	FXString	basicPath;
	FXString	chargenPath;
} sidConfig_t;

/*****************************************************************/

#include "configFrame.h"

class sidConfig : public configFrame {
	FXDECLARE(sidConfig)

private:
	FXDataTarget		_fastSampling;
	FXDataTarget		_bias;
	FXDataTarget		_filter6581Curve;
	FXDataTarget		_filter8580Curve;

	FXVerticalFrame*	_biasFrame;
	FXVerticalFrame*	_filterCurveFrame;
	FXDataTarget		_forceC64Model;
        FXDataTarget            _forceSidModel;
	FXDataTarget		_filter;
	FXDataTarget		_hvscDir;
	FXDataTarget		_kernalDir;
	FXDataTarget		_basicDir;
	FXDataTarget		_chargenDir;

private:
	sidConfig() {};
	sidConfig(const sidConfig&);
	sidConfig& operator=(const sidConfig&);

public:
	sidConfig(FXPacker* win);
	virtual ~sidConfig() {};

public:
	enum{
		ID_FREQUENCY=FXDialogBox::ID_LAST,
		ID_CHANNELS,
		ID_ENGINE,
		ID_ENVIRONMENT,
		ID_CLOCK,
		ID_MODEL,
		ID_ADDRESS,
		ID_HVSC,
		ID_SAMPLING,
		ID_KERNAL,
		ID_BASIC,
		ID_CHARGEN,
		ID_LAST,
	};

public:
	long onCmdFrequency(FXObject*, FXSelector, void*);
	long onCmdChannels(FXObject*, FXSelector, void*);
	long onCmdSampling(FXObject*, FXSelector, void*);
	long onCmdEngine(FXObject*, FXSelector, void*);
	long onCmdClock(FXObject*, FXSelector, void*);
	long onCmdModel(FXObject*, FXSelector, void*);
	long onCmdAddress(FXObject*, FXSelector, void*);
	long onCmdHvsc(FXObject*, FXSelector, void*);
	long onCmdRom(FXObject*, FXSelector, void*);
};

/*****************************************************************/

class sidBackend : public inputBackend {

	friend class sidConfig;

private:
	static sidConfig_t	_settings;
	static const char	_engines[3][8];

private:
	sidplayfp	*_sidplayfp;

	STIL		*_stil;

	SidDatabase	*_db;
	FXuint		_length;
	char		_md5[SidTune::MD5_LENGTH+1];

	SidTune		*_tune;

private:
	sidBackend();

	void loadTune(const int num);

	void openHvsc(const FXString& hvscPath);

	const unsigned char* loadRom(const FXString& romPath);

public:
	~sidBackend();

	static const char name[];

	/// Factory method
	static input* factory() { return new sidBackend(); }

	/// Check if we support ext
	static bool supports(const char* fileName);

	/// Get supported extension
	FXString ext() const;

	/// Open file
	bool open(const FXString& fileName);

	/// Close file
	void close();

	/// Rewind to start
	bool rewind();

	/// Number of subtunes
	unsigned int subtunes() const { return _tune?_tune->getInfo()->songs():0; }

	/// Current subtune
	unsigned int subtune() const { return _tune?_tune->getInfo()->currentSong():0; }

	/// Change subtune
	bool subtune(const unsigned int i);

	/// Get samplerate
	unsigned int samplerate() const { return _tune?_settings.samplerate:0; }

	/// Get channels
	unsigned int channels() const { return _tune?_settings.channels:0; }

	/// Get precision
	sample_t precision() const { return S16; }

	/// Callback function
	size_t fillBuffer(void* buffer, const size_t bufferSize, const unsigned int seconds);

	/// Open config dialog
	void config(FXPacker* win) { new sidConfig(win); }

	/// Get Music directory
	const FXString getMusicDir(void) { return _settings.hvscPath; }

	void loadSettings();

	void saveSettings();
};

#endif

#endif
