/*
FoxTray, system tray icon for the FOX Toolkit

Copyright (c) 2008, Hendrik Rittich
Copyright (c) 2013, Leandro Nini
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in 
      the documentation and/or other materials provided with the 
      distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "FXTrayIcon.h"

#include "FXW32Icon.h"

#ifndef UNICODE
#  include <string.h>
#endif

namespace FX {

FXDEFMAP(FXTrayIcon) FXTrayIconMap[] = {
    FXMAPFUNC(SEL_MOTION, 0, FXTrayIcon::onEvent)
};

FXIMPLEMENT(FXTrayIcon, FXTopWindow, FXTrayIconMap, ARRAYNUMBER(FXTrayIconMap))

DWORD FXTrayIcon::sTrayIconCount = 1;

FXTrayIcon::FXTrayIcon(FXApp* app, const FXString& text, FXIcon* icon, 
        FXPopup* pup, FXObject* target, FXSelector sel,
        FXuint opts) :
	FXTopWindow(app, text, 0, 0, 0, 0,0,0,0, 0,0,0,0, 0,0),
	mIcon(icon),
	mPup(pup),
	mOpts(opts),
	mTooltip(text)
{
	mWIcon = 0;
	mTrayID = sTrayIconCount;
	sTrayIconCount++;

	setTarget(target);
	setSelector(sel);
}

FXTrayIcon::~FXTrayIcon() {
	if (mWIcon) 
	{
		// Icon entfernen
		NOTIFYICONDATA tray_data;
		ZeroMemory(&tray_data, sizeof(tray_data));
		tray_data.cbSize = sizeof(tray_data);
		tray_data.hWnd = static_cast<HWND>(xid);
		tray_data.uID = mTrayID;
		Shell_NotifyIcon(NIM_DELETE, &tray_data);

		// Icon loeschen
		DestroyIcon(mWIcon);
	}
}

void FXTrayIcon::create() {
	FXTopWindow::create();

	if (mWIcon)
		return;

	mIcon->create();

	// Windows Icon erzeugen
    mWIcon = createMswIcon(mIcon);

    mapToManager();
}

void FXTrayIcon::mapToManager() 
{
	NOTIFYICONDATA tray_data;
    setupNotifyData(&tray_data);

	Shell_NotifyIcon(NIM_ADD, &tray_data);
}

void FXTrayIcon::setupNotifyData(NOTIFYICONDATA* data)
{
	ZeroMemory(data, sizeof(*data));
	data->cbSize = sizeof(*data);
	data->hWnd = static_cast<HWND>(xid);
	data->uID = mTrayID;
	data->hIcon = mWIcon;
	data->uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
	data->uCallbackMessage = WM_MOUSEMOVE; // Change ID

	// Tooltip
#ifdef UNICODE
    utf2ncs(data->szTip, mTooltip.text(), 64);
#else
    strncpy(data->szTip, toAscii(mTooltip).text(), 64);
#endif
}

HICON FXTrayIcon::createMswIcon(FXIcon* icon)
{
	// Windows Icon erzeugen
	icon->restore();

	FXW32Icon wIcon(icon);
	wIcon.create();

	return wIcon.createMswIcon();
}

long FXTrayIcon::onEvent(FXObject* obj, FXSelector, void* ptr)
{
	if (ptr == 0)
		return 0;

	FXEvent* event = static_cast<FXEvent*>(ptr);
	int lParam = event->root_x;
	
	if ((lParam == WM_LBUTTONUP && (mOpts & TRAY_MENU_ON_LEFT)) ||
		(lParam == WM_RBUTTONUP && (mOpts & TRAY_MENU_ON_RIGHT)))
	{
		SetForegroundWindow(static_cast<HWND>(xid));

		POINT p;
		GetCursorPos(&p);
		mPup->popup(0, p.x, p.y);
		mPup->grabKeyboard();
		getApp()->runPopup(mPup);
		mPup->ungrabKeyboard();

		return 1;
	}
	if (target &&
		(lParam == WM_LBUTTONUP && (mOpts & TRAY_CMD_ON_LEFT)) ||
		(lParam == WM_RBUTTONUP && (mOpts & TRAY_CMD_ON_RIGHT)))
	{
		return target->tryHandle(obj, FXSEL(SEL_COMMAND, message), ptr);
	}
	return 0;
}

void FXTrayIcon::setIcon(FXIcon* icon)
{
    mIcon = icon;

    if (!mWIcon)
        return; // not created

    // Icon loeschen
	DestroyIcon(mWIcon);

    mWIcon = createMswIcon(icon);

    NOTIFYICONDATA tray_data;
    setupNotifyData(&tray_data);
    Shell_NotifyIcon(NIM_MODIFY, &tray_data);
}

void FXTrayIcon::setText(const FXString& text)
{
    mTooltip = text;

    if (!mWIcon)
        return; // not created

    NOTIFYICONDATA tray_data;
    setupNotifyData(&tray_data);
	Shell_NotifyIcon(NIM_MODIFY, &tray_data);
}

}
